; Echtzeit-Mandelbrotberechnung (320x200)
; Iterationstiefe derzeit 40h
; Autor: Th. Niederhausen
;---------------------------------------
; Ziel: Echtzeit Zooming in Mandelbrotmenge durch ZOOM/ROTATOR jedoch
; mit umgekehrter Zuordnung (Bild->Bildschirm), die vollstndig durch
; eine vorberechnete Quell-/Zieloffset-Tabelle durchgefhrt werden kann.
; (Tabelle wird zu Laufzeit angelegt). Die herausgefallenen Pixel (stets
; gleiche Offsetposition) lassen sich auch tabellieren (bei kleinen Zoom-
; fortschritt sind dies sehr wenig Pixel!) und nur diese Pixel mssen durch
; Mandelbrot-Proc neu durch Iteration berechnet werden! ->Kaum Rechenaufwand,
; -> noch gengend Zeit fr gnadenlose, grandiose, megaaffengeile,
; hypergalaktische Hammereffekte!
; -------------------------------------
; Version 1.A.0 - Berechnung eines kompletten Mandelbrotes und dessen Anzeige
; Qudrate whrend Iteration noch mit IMUL!!!
; Assembliert: 30.3.96

.386c

code    segment byte use16
        assume  cs:code
        org     100h

entry:  jmp     anfang

; berechnet Quadrat-Tabelle in DS - 64K (DWORD) = 16384 Eintrge
; Index ber signed Offset (3 Bit VK, 11 Bit NK)*4 (wegen QWORD)
; Rckgabewert: signed QWORD (1 WORD VK, 1 WORD NK)
qtab    proc
        xor     eax,eax
        mov     si,ax           ; SI=Index auf Tabelle - Anfangswert=0
        mov     edx,eax
        inc     dx              ; mov edx,1 - nur weniger Bytes
        mov     cx,8192
qtabloop:
        mov     ebx,eax
        shr     ebx,6           ; Komma-Korrektur auf 1 WORD NK
        add     eax,edx         ; bischen Mathematik -> nchster Wert berechnet sich ber Steigung:
        add     edx,2           ; f' von x ist 2x -> also immer 2 mehr addieren (in edx=2x)
        mov     [si],ebx        ; Wert speichern
        not     si              ; und auch fr negativen Index nochmal abspeichern
        mov     [si+1],ebx      ; (+1 entspricht Maske um Bit 0-1 auszuschalten (wegenb DWORD); und 0=Mittelpunkt!)
        not     si              ; wieder alten Wert
        add     si,4            ; nchsten Wert maskieren
loop    qtabloop
        ret
qtab    endp

; Definiert Farbeintrge 1-64
defcol  proc
        mov     al,1
        mov     dx,3C8h
        out     dx,al           ; ab Farbeintrag 1 neu definieren
        inc     dx              ; Datenregister des DAC
        mov     ax,64
colorloop:
        out     dx,al           ; Rot
        xchg    al,ah
        out     dx,al           ; Grn
        out     dx,al           ; Blau
        xchg    al,ah
        dec     al
        jnz     colorloop       ; alle Farben durchlaufen
        ret
defcol  endp

; Berechnung: 1 word Vorkomma, 1 word Nachkomma
; esi = c.real; ebx = c.imag    - gendert um Indexzugriffe zu vereinfachen
; Anzeigebereich : -2-2i - 2+2i
; Mandelbrotformel : f(z)=z+c  ; z[0]=0+0i  ; c=x+yi
mandel  proc
;        mov     c_im,0fffdffffh ; y=-2
        mov     ax,0A000h       ; Videosegment
        mov     es,ax
        xor     di,di           ; ES:[DI] auf Videospeicher
        mov     dx,0fffch       ; AND-Maske fr Bit 0-1 und E ausschalten
y_loop:
        mov     dword ptr cs:[c_real+3],0fffdffffh ; x=-2
x_loop:
        mov     esi,cs:dword ptr [c_real+3] ; 1. Iteration: z.re:=x
        mov     ebx,cs:dword ptr [c_imag+3] ; 1. Iteration: z.im:=y
        mov     al,03fh         ; Iterationstiefe
i_loop:
        sar     esi,3           ; Nachkomma-Korrektur (11 bit NK, 3 bit VK)
        sar     ebx,3           ; Nachkomma-Korrektur (11 bit NK, 3 bit VK)
        mov     ebp,esi
        add     ebp,ebx         ; BP=(x+y); 11 bit NK, 3 bit VK
        and     si,dx           ; bit 0-1 und E ausschalten
        mov     esi,[si]        ; aus QTab holen; ESI=x; 1 WORD NK, 1 WORD VK
        and     bl,dl           ; bit 0-1 und E ausschalten
        mov     ebx,[bx]        ; aus QTab holen; EBX=y; 1 WORD NK, 1 WORD VK
        mov     ecx,ebx
        add     ecx,esi         ; ECX=x+y
        cmp     ecx,40000h      ; auf berlauf prfen
        jae     raus            ; berlauf -> Ende der Iteration

        sub     esi,ebx         ; ESI=(z.re)-(z.im)
c_real:
        add     esi,0fffdffffh  ; x addieren -> neues Z.RE; Wert wird modifiziert!
        and     bp,dx
        mov     ebx,ds:[bp]     ; aus QTab holen; EBX=(x+y)=x+y+2xy; 1 WORD NK, 1 WORD VK
        sub     ebx,ecx         ; EBX=2xy
c_imag:
        add     ebx,0fffdffffh  ; y addieren -> neues Z.IM; Wert wird modifiziert!
        dec     al
        jz      raus            ; maximale Iterationstiefe erreicht?
        jmp     i_loop          ; nein! -> nchste Iteration
raus:
        inc     al              ; keine Farbe 0 verwenden
        stosb                   ; Pixel ausgeben
        cmp     dword ptr cs:[c_real+3],1fdcbh ; maximale x-Position
        jge     weiter_y        ; ja -> nchstes Y
        add     dword ptr cs:[c_real+3],334h ; c.re (x) erhhen
        jmp     x_loop
weiter_y:
        cmp     di,64000        ; letzter Punkt schon berechnet?
        jae     fertig          ; ja -> fertig
        add     dword ptr cs:[c_imag+3],51fh       ; c.im (y) erhhen
        jmp     y_loop
fertig:
        ret
mandel  endp

anfang:
        cli
        mov     ax,cs
        mov     ss,ax
        mov     sp,(offset done)+500 ; Stack erstellen
        sti

        mov     ah,4Ah
        push    cs
        pop     es
        mov     bx,(offset done)+520 ; Sicherheitsbereich wegen Runden
        shr     bx,4            ; Angabe in Paragraphen
        int     21h             ; unbentigten Speicher freigeben
        jc      ende            ; Fehler!

        mov     ah,48h
        mov     bx,4096
        int     21h             ; ein Segment allokieren
        jc      ende            ; zu wenig Speicher
        mov     ds,ax           ; Datensegment fr QTab

        mov     ax,13h
        int     10h             ; linearer Grafikmodus 13h (320x200x256)

        call    qtab            ; generiert Quadrat-Tabelle (64K in DS)
;        call    defcol          ; definiert Farbeintrge 1 bis 64
        call    mandel          ; berechnet Mandelbrotmenge und zeigt diese an

        mov     ah,10h
        int     16h             ; auf Taste warten

        push    ds
        pop     es
        mov     ah,49h
        int     21h             ; Segment fr QTab wieder freigeben
ende:
        mov     ax,3
        int     10h             ; wieder Textmodus
        mov     ax,4C00h
        int     21h             ; beenden ohne Fehlercode

done:
code    ends
        end     entry
